/*
 * kgraft_patch_bsc986377
 *
 * Fix for CVE-2016-4997, bsc#986377, ip_tables module
 *
 *  Upstream commits:
 *  ce683e5f9d04 ("netfilter: x_tables: check for bogus target offset")
 *  ... and more
 *
 *  SLE12 commits:
 *  c432d109f6a4f68a92a623e3968e8d51f5a47133
 *  ... (>10 patches in the series)
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <net/sock.h>

#include "kgr_patch_bsc986377_xtables.h"

#define KGR_PATCHED_MODULE      "ip_tables"

static int (*kgr_check_match)(struct xt_entry_match *m, struct xt_mtchk_param *par);
static int (*kgr_check_target)(struct ipt_entry *e, struct net *net, const char *name);
static void (*kgr_cleanup_match)(struct xt_entry_match *m, struct net *net);
static void (*kgr_cleanup_entry)(struct ipt_entry *e, struct net *net);
static void (*kgr_compat_release_entry)(struct compat_ipt_entry *e);
static int (*kgr_do_add_counters)(struct net *net, const void __user *user, unsigned int len, int compat);
static int (*kgr__do_replace)(struct net *net, const char *name, unsigned int valid_hooks, struct xt_table_info *newinfo, unsigned int num_counters, void __user *counters_ptr);
static void (*kgr_xt_compat_lock)(u_int8_t af);
static int (*kgr_xt_compat_match_offset)(const struct xt_match *match);
static int (*kgr_xt_compat_target_offset)(const struct xt_target *target);
static void (*kgr_xt_compat_flush_offsets)(u_int8_t af);
static void (*kgr_xt_compat_init_offsets)(u_int8_t af, unsigned int number);
static int (*kgr_xt_compat_add_offset)(u_int8_t af, unsigned int offset, int delta);
static struct xt_match *(*kgr_xt_request_find_match)(uint8_t nfproto, const char *name, uint8_t revision);
static struct xt_target *(*kgr_xt_request_find_target)(u8 af, const char *name, u8 revision);
static struct xt_table_info *(*kgr_xt_alloc_table_info)(unsigned int size);
static void (*kgr_xt_free_table_info)(struct xt_table_info *info);
static void (*kgr_xt_compat_unlock)(u_int8_t af);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "ip_tables:check_match", (void *)&kgr_check_match },
	{ "ip_tables:check_target", (void *)&kgr_check_target },
	{ "ip_tables:cleanup_match", (void *)&kgr_cleanup_match },
	{ "ip_tables:cleanup_entry", (void *)&kgr_cleanup_entry },
	{ "ip_tables:compat_release_entry", (void *)&kgr_compat_release_entry },
	{ "ip_tables:do_add_counters", (void *)&kgr_do_add_counters },
	{ "ip_tables:__do_replace", (void *)&kgr__do_replace },
	{ "xt_compat_lock", (void *)&kgr_xt_compat_lock },
	{ "xt_compat_match_offset", (void *)&kgr_xt_compat_match_offset },
	{ "xt_compat_target_offset", (void *)&kgr_xt_compat_target_offset },
	{ "xt_compat_flush_offsets", (void *)&kgr_xt_compat_flush_offsets },
	{ "xt_compat_init_offsets", (void *)&kgr_xt_compat_init_offsets },
	{ "xt_compat_add_offset", (void *)&kgr_xt_compat_add_offset },
	{ "xt_request_find_match", (void *)&kgr_xt_request_find_match },
	{ "xt_request_find_target", (void *)&kgr_xt_request_find_target },
	{ "xt_alloc_table_info", (void *)&kgr_xt_alloc_table_info },
	{ "xt_free_table_info", (void *)&kgr_xt_free_table_info },
	{ "xt_compat_unlock", (void *)&kgr_xt_compat_unlock },
};

#ifdef DEBUG_IP_FIREWALL_USER
#define duprintf(format, args...) pr_info(format , ## args)
#else
#define duprintf(format, args...)
#endif

struct compat_ipt_replace {
	char			name[XT_TABLE_MAXNAMELEN];
	u32			valid_hooks;
	u32			num_entries;
	u32			size;
	u32			hook_entry[NF_INET_NUMHOOKS];
	u32			underflow[NF_INET_NUMHOOKS];
	u32			num_counters;
	compat_uptr_t		counters;	/* struct xt_counters * */
	struct compat_ipt_entry	entries[0];
};

/* inlined */
static inline bool kgr_unconditional(const struct ipt_entry *e)
{
	static const struct ipt_ip uncond;

	return e->target_offset == sizeof(struct ipt_entry) &&
	       memcmp(&e->ip, &uncond, sizeof(uncond)) == 0;
#undef FWINV
}

/* inlined */
static inline const struct xt_entry_target *
kgr_ipt_get_target_c(const struct ipt_entry *e)
{
	return ipt_get_target((struct ipt_entry *)e);
}

/* patched function */
/*
 * Fix CVE-2016-4997
 *  prototype changed
 */
int kgr_ip_mark_source_chains(const struct xt_table_info *newinfo,
/*
 *		   unsigned int valid_hooks, void *entry0)
 */
			unsigned int valid_hooks, void *entry0,
			unsigned int *offsets)
{
	unsigned int hook;

	/* No recursion; use packet counter to save back ptrs (reset
	   to 0 as we leave), and comefrom to save source hook bitmask */
	for (hook = 0; hook < NF_INET_NUMHOOKS; hook++) {
		unsigned int pos = newinfo->hook_entry[hook];
		struct ipt_entry *e = (struct ipt_entry *)(entry0 + pos);

		if (!(valid_hooks & (1 << hook)))
			continue;

		/* Set initial back pointer. */
		e->counters.pcnt = pos;

		for (;;) {
			const struct xt_standard_target *t
				= (void *)kgr_ipt_get_target_c(e);
			int visited = e->comefrom & (1 << hook);

			if (e->comefrom & (1 << NF_INET_NUMHOOKS)) {
				pr_err("iptables: loop hook %u pos %u %08X.\n",
				       hook, pos, e->comefrom);
				return 0;
			}
			e->comefrom |= ((1 << hook) | (1 << NF_INET_NUMHOOKS));

			/* Unconditional return/END. */
			if ((kgr_unconditional(e) &&
			     (strcmp(t->target.u.user.name,
				     XT_STANDARD_TARGET) == 0) &&
			     t->verdict < 0) || visited) {
				unsigned int oldpos, size;

				if ((strcmp(t->target.u.user.name,
					    XT_STANDARD_TARGET) == 0) &&
				    t->verdict < -NF_MAX_VERDICT - 1) {
					duprintf("mark_source_chains: bad "
						"negative verdict (%i)\n",
								t->verdict);
					return 0;
				}

				/* Return: backtrack through the last
				   big jump. */
				do {
					e->comefrom ^= (1<<NF_INET_NUMHOOKS);
#ifdef DEBUG_IP_FIREWALL_USER
					if (e->comefrom
					    & (1 << NF_INET_NUMHOOKS)) {
						duprintf("Back unset "
							 "on hook %u "
							 "rule %u\n",
							 hook, pos);
					}
#endif
					oldpos = pos;
					pos = e->counters.pcnt;
					e->counters.pcnt = 0;

					/* We're at the start. */
					if (pos == oldpos)
						goto next;

					e = (struct ipt_entry *)
						(entry0 + pos);
				} while (oldpos == pos + e->next_offset);

				/* Move along one */
				size = e->next_offset;
				e = (struct ipt_entry *)
					(entry0 + pos + size);
				/*
				 * Fix CVE-2016-4997
				 *  +2 lines
				 */
				if (pos + size >= newinfo->size)
					return 0;
				e->counters.pcnt = pos;
				pos += size;
			} else {
				int newpos = t->verdict;

				if (strcmp(t->target.u.user.name,
					   XT_STANDARD_TARGET) == 0 &&
				    newpos >= 0) {
					if (newpos > newinfo->size -
						sizeof(struct ipt_entry)) {
						duprintf("mark_source_chains: "
							"bad verdict (%i)\n",
								newpos);
						return 0;
					}
					/* This a jump; chase it. */
					duprintf("Jump rule %u -> %u\n",
						 pos, newpos);
					/*
					 * Fix CVE-2016-4997
					 *  +5 lines
					 */
					if (!kgr_xt_find_jump_offset(offsets, newpos,
								 newinfo->number))
						return 0;
					e = (struct ipt_entry *)
						(entry0 + newpos);
				} else {
					/* ... this is a fallthru */
					newpos = pos + e->next_offset;
					/*
					 * Fix CVE-2016-4997
					 *  +2 lines
					 */
					if (newpos >= newinfo->size)
						return 0;
				}
				e = (struct ipt_entry *)
					(entry0 + newpos);
				e->counters.pcnt = pos;
				pos = newpos;
			}
		}
		next:
		duprintf("Finished chain %u\n", hook);
	}
	return 1;
}

/* inlined */
static bool kgr_check_underflow(const struct ipt_entry *e)
{
	const struct xt_entry_target *t;
	unsigned int verdict;

	if (!kgr_unconditional(e))
		return false;
	t = kgr_ipt_get_target_c(e);
	if (strcmp(t->u.user.name, XT_STANDARD_TARGET) != 0)
		return false;
	verdict = ((struct xt_standard_target *)t)->verdict;
	verdict = -verdict - 1;
	return verdict == NF_DROP || verdict == NF_ACCEPT;
}

/* inlined */
static bool
kgr_ip_checkentry(const struct ipt_ip *ip)
{
	if (ip->flags & ~IPT_F_MASK) {
		duprintf("Unknown flag bits set: %08X\n",
			 ip->flags & ~IPT_F_MASK);
		return false;
	}
	if (ip->invflags & ~IPT_INV_MASK) {
		duprintf("Unknown invflag bits set: %08X\n",
			 ip->invflags & ~IPT_INV_MASK);
		return false;
	}
	return true;
}

/* inlined */
static int
kgr_check_entry_size_and_hooks(struct ipt_entry *e,
			   struct xt_table_info *newinfo,
			   const unsigned char *base,
			   const unsigned char *limit,
			   const unsigned int *hook_entries,
			   const unsigned int *underflows,
			   unsigned int valid_hooks)
{
	unsigned int h;
	int err;

	if ((unsigned long)e % __alignof__(struct ipt_entry) != 0 ||
	    (unsigned char *)e + sizeof(struct ipt_entry) >= limit ||
	    (unsigned char *)e + e->next_offset > limit) {
		duprintf("Bad offset %p\n", e);
		return -EINVAL;
	}

	if (e->next_offset
	    < sizeof(struct ipt_entry) + sizeof(struct xt_entry_target)) {
		duprintf("checking: element %p size %u\n",
			 e, e->next_offset);
		return -EINVAL;
	}

	/*
	 * Fix CVE-2016-4997
	 *  -err = check_entry(e);
	 *  +5 lines
	 */
	if (!kgr_ip_checkentry(&e->ip))
		return -EINVAL;

	err = kgr_xt_check_entry_offsets(e, e->elems, e->target_offset,
				     e->next_offset);
	if (err)
		return err;

	/* Check hooks & underflows */
	for (h = 0; h < NF_INET_NUMHOOKS; h++) {
		if (!(valid_hooks & (1 << h)))
			continue;
		if ((unsigned char *)e - base == hook_entries[h])
			newinfo->hook_entry[h] = hook_entries[h];
		if ((unsigned char *)e - base == underflows[h]) {
			if (!kgr_check_underflow(e)) {
				pr_debug("Underflows must be unconditional and "
					 "use the STANDARD target with "
					 "ACCEPT/DROP\n");
				return -EINVAL;
			}
			newinfo->underflow[h] = underflows[h];
		}
	}

	/* Clear counters and comefrom */
	e->counters = ((struct xt_counters) { 0, 0 });
	e->comefrom = 0;
	return 0;
}

/* inlined */
static int
kgr_find_check_match(struct xt_entry_match *m, struct xt_mtchk_param *par)
{
	struct xt_match *match;
	int ret;

	match = kgr_xt_request_find_match(NFPROTO_IPV4, m->u.user.name,
				      m->u.user.revision);
	if (IS_ERR(match)) {
		duprintf("find_check_match: `%s' not found\n", m->u.user.name);
		return PTR_ERR(match);
	}
	m->u.kernel.match = match;

	ret = kgr_check_match(m, par);
	if (ret)
		goto err;

	return 0;
err:
	module_put(m->u.kernel.match->me);
	return ret;
}

/* inlined */
static int
kgr_find_check_entry(struct ipt_entry *e, struct net *net, const char *name,
		 unsigned int size)
{
	struct xt_entry_target *t;
	struct xt_target *target;
	int ret;
	unsigned int j;
	struct xt_mtchk_param mtpar;
	struct xt_entry_match *ematch;

	j = 0;
	mtpar.net	= net;
	mtpar.table     = name;
	mtpar.entryinfo = &e->ip;
	mtpar.hook_mask = e->comefrom;
	mtpar.family    = NFPROTO_IPV4;
	xt_ematch_foreach(ematch, e) {
		ret = kgr_find_check_match(ematch, &mtpar);
		if (ret != 0)
			goto cleanup_matches;
		++j;
	}

	t = ipt_get_target(e);
	target = kgr_xt_request_find_target(NFPROTO_IPV4, t->u.user.name,
					t->u.user.revision);
	if (IS_ERR(target)) {
		duprintf("find_check_entry: `%s' not found\n", t->u.user.name);
		ret = PTR_ERR(target);
		goto cleanup_matches;
	}
	t->u.kernel.target = target;

	ret = kgr_check_target(e, net, name);
	if (ret)
		goto err;
	return 0;
 err:
	module_put(t->u.kernel.target->me);
 cleanup_matches:
	xt_ematch_foreach(ematch, e) {
		if (j-- == 0)
			break;
		kgr_cleanup_match(ematch, net);
	}
	return ret;
}

/* patched function */
int kgr_ip_translate_table(struct net *net, struct xt_table_info *newinfo, void *entry0,
                const struct ipt_replace *repl)
{
	struct ipt_entry *iter;
	/*
	 * Fix CVE-2016-4997
	 *  +1 line
	 */
	unsigned int *offsets;
	unsigned int i;
	int ret = 0;

	newinfo->size = repl->size;
	newinfo->number = repl->num_entries;

	/* Init all hooks to impossible value. */
	for (i = 0; i < NF_INET_NUMHOOKS; i++) {
		newinfo->hook_entry[i] = 0xFFFFFFFF;
		newinfo->underflow[i] = 0xFFFFFFFF;
	}

	duprintf("translate_table: size %u\n", newinfo->size);
	/*
	 * Fix CVE-2016-4997
	 *  +3 lines
	 */
	offsets = kgr_xt_alloc_entry_offsets(newinfo->number);
	if (!offsets)
		return -ENOMEM;
	i = 0;
	/* Walk through entries, checking offsets. */
	xt_entry_foreach(iter, entry0, newinfo->size) {
		ret = kgr_check_entry_size_and_hooks(iter, newinfo, entry0,
						 entry0 + repl->size,
						 repl->hook_entry,
						 repl->underflow,
						 repl->valid_hooks);
		if (ret != 0)
		/*
		 * Fix CVE-2016-4997
		 *  -	return ret;
		 *  +3 lines
		 */
			goto out_free;
		if (i < repl->num_entries)
			offsets[i] = (void *)iter - entry0;
		++i;
		if (strcmp(ipt_get_target(iter)->u.user.name,
		    XT_ERROR_TARGET) == 0)
			++newinfo->stacksize;
	}

	/*
	 * Fix CVE-2016-4997
	 *  +1 line
	 */
	ret = -EINVAL;
	if (i != repl->num_entries) {
		duprintf("translate_table: %u not %u entries\n",
			 i, repl->num_entries);
		/*
		 * Fix CVE-2016-4997
		 *  -return -EINVAL;
		 *  +1 line
		 */
		goto out_free;
	}

	/* Check hooks all assigned */
	for (i = 0; i < NF_INET_NUMHOOKS; i++) {
		/* Only hooks which are valid */
		if (!(repl->valid_hooks & (1 << i)))
			continue;
		if (newinfo->hook_entry[i] == 0xFFFFFFFF) {
			duprintf("Invalid hook entry %u %u\n",
				 i, repl->hook_entry[i]);
			/*
			 * Fix CVE-2016-4997
			 *  -return -EINVAL;
			 *  +1 line
			 */
			goto out_free;
		}
		if (newinfo->underflow[i] == 0xFFFFFFFF) {
			duprintf("Invalid underflow %u %u\n",
				 i, repl->underflow[i]);
			/*
			 * Fix CVE-2016-4997
			 *  -return -EINVAL;
			 *  +1 line
			 */
			goto out_free;
		}
	}

	/*
	 * Fix CVE-2016-4997
	 *  -if (!kgr_ip_mark_source_chains(newinfo, repl->valid_hooks, entry0))
	 *  -	return -ELOOP;
	 *  +5 lines
	 */
	if (!kgr_ip_mark_source_chains(newinfo, repl->valid_hooks, entry0, offsets)) {
		ret = -ELOOP;
		goto out_free;
	}
	kvfree(offsets);

	/* Finally, each sanity check must pass */
	i = 0;
	xt_entry_foreach(iter, entry0, newinfo->size) {
		ret = kgr_find_check_entry(iter, net, repl->name, repl->size);
		if (ret != 0)
			break;
		++i;
	}

	if (ret != 0) {
		xt_entry_foreach(iter, entry0, newinfo->size) {
			if (i-- == 0)
				break;
			kgr_cleanup_entry(iter, net);
		}
		return ret;
	}

	/* And one copy for every other CPU */
	for_each_possible_cpu(i) {
		if (newinfo->entries[i] && newinfo->entries[i] != entry0)
			memcpy(newinfo->entries[i], entry0, newinfo->size);
	}

	return ret;
/*
 * Fix CVE-2016-4997
 *  +3 lines
 */
 out_free:
	kvfree(offsets);
	return ret;
}

/* inlined */
/*
 * Fix CVE-2016-4997
 *  prototype changed
 */
static int
kgr_compat_find_calc_match(struct xt_entry_match *m,
/*
 *		       const char *name,
 */
		       const struct ipt_ip *ip,
		       unsigned int hookmask,
		       int *size)
{
	struct xt_match *match;

	match = kgr_xt_request_find_match(NFPROTO_IPV4, m->u.user.name,
				      m->u.user.revision);
	if (IS_ERR(match)) {
		duprintf("compat_check_calc_match: `%s' not found\n",
			 m->u.user.name);
		return PTR_ERR(match);
	}
	m->u.kernel.match = match;
	*size += kgr_xt_compat_match_offset(match);
	return 0;
}

/* inlined */
/*
 * Fix CVE-2016-4997
 *  prototype changed
 */
static int
kgr_check_compat_entry_size_and_hooks(struct compat_ipt_entry *e,
				  struct xt_table_info *newinfo,
				  unsigned int *size,
				  const unsigned char *base,
/*
 *				  const unsigned char *limit,
 *				  const unsigned int *hook_entries,
 *				  const unsigned int *underflows,
 *				  const char *name)
 */
				  const unsigned char *limit)
{
	struct xt_entry_match *ematch;
	struct xt_entry_target *t;
	struct xt_target *target;
	unsigned int entry_offset;
	unsigned int j;
	/*
	 * Fix CVE-2016-4997
	 *  -int ret, off, h;
	 *  +1 line
	 */
	int ret, off;

	duprintf("check_compat_entry_size_and_hooks %p\n", e);
	if ((unsigned long)e % __alignof__(struct compat_ipt_entry) != 0 ||
	    (unsigned char *)e + sizeof(struct compat_ipt_entry) >= limit ||
	    (unsigned char *)e + e->next_offset > limit) {
		duprintf("Bad offset %p, limit = %p\n", e, limit);
		return -EINVAL;
	}

	if (e->next_offset < sizeof(struct compat_ipt_entry) +
			     sizeof(struct compat_xt_entry_target)) {
		duprintf("checking: element %p size %u\n",
			 e, e->next_offset);
		return -EINVAL;
	}

	/*
	 * Fix CVE-2016-4997
	 *  -* For purposes of check_entry casting the compat entry is fine *
	 *  -ret = check_entry((struct ipt_entry *)e);
	 *  +5 lines
	 */
	if (!kgr_ip_checkentry(&e->ip))
		return -EINVAL;

	ret = kgr_xt_compat_check_entry_offsets(e, e->elems,
					    e->target_offset, e->next_offset);
	if (ret)
		return ret;

	off = sizeof(struct ipt_entry) - sizeof(struct compat_ipt_entry);
	entry_offset = (void *)e - (void *)base;
	j = 0;
	xt_ematch_foreach(ematch, e) {
		/*
		 * Fix CVE-2016-4997
		 *  -ret = kgr_compat_find_calc_match(ematch, name,
		 *  -			     &e->ip, e->comefrom, &off);
		 *  +1 line
		 */
		ret = kgr_compat_find_calc_match(ematch, &e->ip, e->comefrom, &off);
		if (ret != 0)
			goto release_matches;
		++j;
	}

	t = compat_ipt_get_target(e);
	target = kgr_xt_request_find_target(NFPROTO_IPV4, t->u.user.name,
					t->u.user.revision);
	if (IS_ERR(target)) {
		duprintf("check_compat_entry_size_and_hooks: `%s' not found\n",
			 t->u.user.name);
		ret = PTR_ERR(target);
		goto release_matches;
	}
	t->u.kernel.target = target;

	off += kgr_xt_compat_target_offset(target);
	*size += off;
	ret = kgr_xt_compat_add_offset(AF_INET, entry_offset, off);
	if (ret)
		goto out;

	/*
	 * Fix CVE-2016-4997
	 *  -* Check hooks & underflows *
	 *  -for (h = 0; h < NF_INET_NUMHOOKS; h++) {
	 *  -	if ((unsigned char *)e - base == hook_entries[h])
	 *  -		newinfo->hook_entry[h] = hook_entries[h];
	 *  -	if ((unsigned char *)e - base == underflows[h])
	 *  -		newinfo->underflow[h] = underflows[h];
	 *  -}

	 *  -* Clear counters and comefrom *
	 *  -memset(&e->counters, 0, sizeof(e->counters));
	 *  -e->comefrom = 0;
	 */
	return 0;

out:
	module_put(t->u.kernel.target->me);
release_matches:
	xt_ematch_foreach(ematch, e) {
		if (j-- == 0)
			break;
		module_put(ematch->u.kernel.match->me);
	}
	return ret;
}

/* inlined */
/*
 * Fix CVE-2016-4997
 *  prototype changed
 */
static void
kgr_compat_copy_entry_from_user(struct compat_ipt_entry *e, void **dstptr,
/*
 *			    unsigned int *size, const char *name,
 */
			    unsigned int *size,
			    struct xt_table_info *newinfo, unsigned char *base)
{
	struct xt_entry_target *t;
	struct xt_target *target;
	struct ipt_entry *de;
	unsigned int origsize;
	/*
	 * Fix CVE-2016-4997
	 *  -int ret, h;
	 *  +1 line
	 */
	int h;
	struct xt_entry_match *ematch;

	/*
	 * Fix CVE-2016-4997
	 *  -ret = 0;
	 */
	origsize = *size;
	de = (struct ipt_entry *)*dstptr;
	memcpy(de, e, sizeof(struct ipt_entry));
	memcpy(&de->counters, &e->counters, sizeof(e->counters));

	*dstptr += sizeof(struct ipt_entry);
	*size += sizeof(struct ipt_entry) - sizeof(struct compat_ipt_entry);

	/*
	 * Fix CVE-2016-4997
	 *  -xt_ematch_foreach(ematch, e) {
	 *  -	ret = xt_compat_match_from_user(ematch, dstptr, size);
	 *  -	if (ret != 0)
	 *  -		return ret;
	 *  -}
	 *  +3 lines
	 */
	xt_ematch_foreach(ematch, e)
		kgr_xt_compat_match_from_user(ematch, dstptr, size);

	de->target_offset = e->target_offset - (origsize - *size);
	t = compat_ipt_get_target(e);
	target = t->u.kernel.target;
	kgr_xt_compat_target_from_user(t, dstptr, size);

	de->next_offset = e->next_offset - (origsize - *size);
	for (h = 0; h < NF_INET_NUMHOOKS; h++) {
		if ((unsigned char *)de - base < newinfo->hook_entry[h])
			newinfo->hook_entry[h] -= origsize - *size;
		if ((unsigned char *)de - base < newinfo->underflow[h])
			newinfo->underflow[h] -= origsize - *size;
	}
	/*
	 * Fix CVE-2016-4997
	 *  -return ret;
	 */
}

/* inlined */
/*
 * Fix CVE-2016-4997
 *  function removed
 */
//static int
//kgr_compat_check_entry(struct ipt_entry *e, struct net *net, const char *name)
//{
//	struct xt_entry_match *ematch;
//	struct xt_mtchk_param mtpar;
//	unsigned int j;
//	int ret = 0;
//
//	j = 0;
//	mtpar.net	= net;
//	mtpar.table     = name;
//	mtpar.entryinfo = &e->ip;
//	mtpar.hook_mask = e->comefrom;
//	mtpar.family    = NFPROTO_IPV4;
//	xt_ematch_foreach(ematch, e) {
//		ret = kgr_check_match(ematch, &mtpar);
//		if (ret != 0)
//			goto cleanup_matches;
//		++j;
//	}
//
//	ret = kgr_check_target(e, net, name);
//	if (ret)
//		goto cleanup_matches;
//	return 0;
//
// cleanup_matches:
//	xt_ematch_foreach(ematch, e) {
//		if (j-- == 0)
//			break;
//		kgr_cleanup_match(ematch, net);
//	}
//	return ret;
//}

/* patched function */
/*
 * Fix CVE-2016-4997
 *  prototype changed
 */
int kgr_ip_translate_compat_table(struct net *net,
/*
 *		       const char *name,
 *		       unsigned int valid_hooks,
 */
		       struct xt_table_info **pinfo,
		       void **pentry0,
/*
 *		       unsigned int total_size,
 *		       unsigned int number,
 *		       unsigned int *hook_entries,
 *		       unsigned int *underflows)
 */
		       const struct compat_ipt_replace *compatr)
{
	unsigned int i, j;
	struct xt_table_info *newinfo, *info;
	void *pos, *entry0, *entry1;
	struct compat_ipt_entry *iter0;
	/*
	 * Fix CVE-2016-4997
	 *  -struct ipt_entry *iter1;
	 *  +1 line
	 */
	struct ipt_replace repl;
	unsigned int size;
	int ret;

	info = *pinfo;
	entry0 = *pentry0;
	/*
	 * Fix CVE-2016-4997
	 *  -size = total_size;
	 *  -info->number = number;
	 *
	 *  -* Init all hooks to impossible value. *
	 *  -for (i = 0; i < NF_INET_NUMHOOKS; i++) {
	 *  -	info->hook_entry[i] = 0xFFFFFFFF;
	 *  -	info->underflow[i] = 0xFFFFFFFF;
	 *  -}
	 *  +2 lines
	 */
	size = compatr->size;
	info->number = compatr->num_entries;

	duprintf("translate_compat_table: size %u\n", info->size);
	j = 0;
	kgr_xt_compat_lock(AF_INET);
	/*
	 * Fix CVE-2016-4997
	 *  -xt_compat_init_offsets(AF_INET, number);
	 *  +1 line
	 */
	kgr_xt_compat_init_offsets(AF_INET, compatr->num_entries);
	/* Walk through entries, checking offsets. */
	/*
	 * Fix CVE-2016-4997
	 *  -xt_entry_foreach(iter0, entry0, total_size) {
	 *  +1 line
	 */
	xt_entry_foreach(iter0, entry0, compatr->size) {
		ret = kgr_check_compat_entry_size_and_hooks(iter0, info, &size,
							entry0,
							/*
							 * Fix CVE-2016-4997
							 *  -entry0 + total_size,
							 *  -hook_entries,
							 *  -underflows,
							 *  -name);
							 *  +1 line
							 */
							entry0 + compatr->size);
		if (ret != 0)
			goto out_unlock;
		++j;
	}

	ret = -EINVAL;
	/*
	 * Fix CVE-2016-4997
	 *  -if (j != number) {
	 *  +1 line
	 */
	if (j != compatr->num_entries) {
		duprintf("translate_compat_table: %u not %u entries\n",
			/*
			 * Fix CVE-2016-4997
			 *  -j, number);
			 *  +1 line
			 */
			j, compatr->num_entries);
		goto out_unlock;
	}

	/*
	 * Fix CVE-2016-4997
	 *  -* Check hooks all assigned *
	 *  -for (i = 0; i < NF_INET_NUMHOOKS; i++) {
	 *  -	* Only hooks which are valid *
	 *  -	if (!(valid_hooks & (1 << i)))
	 *  -		continue;
	 *  -	if (info->hook_entry[i] == 0xFFFFFFFF) {
	 *  -		duprintf("Invalid hook entry %u %u\n",
	 *  -			 i, hook_entries[i]);
	 *  -		goto out_unlock;
	 *  -	}
	 *  -	if (info->underflow[i] == 0xFFFFFFFF) {
	 *  -		duprintf("Invalid underflow %u %u\n",
	 *  -			 i, underflows[i]);
	 *  -		goto out_unlock;
	 *  -	}
	 *  -}
	 *  -
	 */
	ret = -ENOMEM;
	newinfo = kgr_xt_alloc_table_info(size);
	if (!newinfo)
		goto out_unlock;

	/*
	 * Fix CVE-2016-4997
	 *  -newinfo->number = number;
	 *  +1 line
	 */
	newinfo->number = compatr->num_entries;
	for (i = 0; i < NF_INET_NUMHOOKS; i++) {
		/*
		 * Fix CVE-2016-4997
		 *  -newinfo->hook_entry[i] = info->hook_entry[i];
		 *  -newinfo->underflow[i] = info->underflow[i];
		 *  +2 lines
		 */
		newinfo->hook_entry[i] = compatr->hook_entry[i];
		newinfo->underflow[i] = compatr->underflow[i];
	}
	entry1 = newinfo->entries[raw_smp_processor_id()];
	pos = entry1;
	/*
	 * Fix CVE-2016-4997
	 *  -size = total_size;
	 *  -xt_entry_foreach(iter0, entry0, total_size) {
	 *  -	ret = kgr_compat_copy_entry_from_user(iter0, &pos, &size,
	 *  -					  name, newinfo, entry1);
	 *  -	if (ret != 0)
	 *  -		break;
	 *  -}
	 *  +12 lines
	 */
	size = compatr->size;
	xt_entry_foreach(iter0, entry0, compatr->size)
		kgr_compat_copy_entry_from_user(iter0, &pos, &size,
					    newinfo, entry1);

	/* all module references in entry0 are now gone.
	 * entry1/newinfo contains a 64bit ruleset that looks exactly as
	 * generated by 64bit userspace.
	 *
	 * Call standard translate_table() to validate all hook_entrys,
	 * underflows, check for loops, etc.
	*/
	kgr_xt_compat_flush_offsets(AF_INET);
	kgr_xt_compat_unlock(AF_INET);
	/*
	 * Fix CVE-2016-4997
	 *  -if (ret)
	 *  -	goto free_newinfo;
	 *
	 *  -ret = -ELOOP;
	 *  -if (!kgr_ip_mark_source_chains(newinfo, valid_hooks, entry1))
	 *  -	goto free_newinfo;
	 *  +1 line
	 */
	memcpy(&repl, compatr, sizeof(*compatr));

	/*
	 * Fix CVE-2016-4997
	 *  -i = 0;
	 *  -xt_entry_foreach(iter1, entry1, newinfo->size) {
	 *  -	ret = kgr_compat_check_entry(iter1, net, name);
	 *  -	if (ret != 0)
	 *  -		break;
	 *  -	++i;
	 *  -	if (strcmp(ipt_get_target(iter1)->u.user.name,
	 *  -	    XT_ERROR_TARGET) == 0)
	 *  -		++newinfo->stacksize;
	 *  -}
	 *  -if (ret) {
	 *  -	*
	 *  -	 * The first i matches need cleanup_entry (calls ->destroy)
	 *  -	 * because they had called ->check already. The other j-i
	 *  -	 * entries need only release.
	 *  -	 *
	 *  -	int skip = i;
	 *  -	j -= i;
	 *  -	xt_entry_foreach(iter0, entry0, newinfo->size) {
	 *  -		if (skip-- > 0)
	 *  -			continue;
	 *  -		if (j-- == 0)
	 *  -			break;
	 *  -		kgr_compat_release_entry(iter0);
	 *  -	}
	 *  -	xt_entry_foreach(iter1, entry1, newinfo->size) {
	 *  -		if (i-- == 0)
	 *  -			break;
	 *  -		kgr_cleanup_entry(iter1, net);
	 *  -	}
	 *  -	xt_free_table_info(newinfo);
	 *  -	return ret;
	 *  -}
	 *  +11 lines
	 */
	for (i = 0; i < NF_INET_NUMHOOKS; i++) {
		repl.hook_entry[i] = newinfo->hook_entry[i];
		repl.underflow[i] = newinfo->underflow[i];
	}

	repl.num_counters = 0;
	repl.counters = NULL;
	repl.size = newinfo->size;
	ret = kgr_ip_translate_table(net, newinfo, entry1, &repl);
	if (ret)
		goto free_newinfo;

	/* And one copy for every other CPU */
	for_each_possible_cpu(i)
		if (newinfo->entries[i] && newinfo->entries[i] != entry1)
			memcpy(newinfo->entries[i], entry1, newinfo->size);

	*pinfo = newinfo;
	*pentry0 = entry1;
	kgr_xt_free_table_info(info);
	return 0;

free_newinfo:
	kgr_xt_free_table_info(newinfo);
/*
 * Fix CVE-2016-4997
 *  -out:
 *  -	xt_entry_foreach(iter0, entry0, total_size) {
 *  +5 lines
 */
	return ret;
out_unlock:
	kgr_xt_compat_flush_offsets(AF_INET);
	kgr_xt_compat_unlock(AF_INET);
	xt_entry_foreach(iter0, entry0, compatr->size) {
		if (j-- == 0)
			break;
		kgr_compat_release_entry(iter0);
	}
	return ret;
/*
 * Fix CVE-2016-4997
 *  -out_unlock:
 *  -	xt_compat_flush_offsets(AF_INET);
 *  -	xt_compat_unlock(AF_INET);
 *  -	goto out;
 */
}

/* isra optimized */
static int
kgr_compat_do_replace(struct net *net, void __user *user, unsigned int len)
{
	int ret;
	struct compat_ipt_replace tmp;
	struct xt_table_info *newinfo;
	void *loc_cpu_entry;
	struct ipt_entry *iter;

	if (copy_from_user(&tmp, user, sizeof(tmp)) != 0)
		return -EFAULT;

	/* overflow check */
	if (tmp.size >= INT_MAX / num_possible_cpus())
		return -ENOMEM;
	if (tmp.num_counters >= INT_MAX / sizeof(struct xt_counters))
		return -ENOMEM;
	tmp.name[sizeof(tmp.name)-1] = 0;

	newinfo = kgr_xt_alloc_table_info(tmp.size);
	if (!newinfo)
		return -ENOMEM;

	/* choose the copy that is on our node/cpu */
	loc_cpu_entry = newinfo->entries[raw_smp_processor_id()];
	if (copy_from_user(loc_cpu_entry, user + sizeof(tmp),
			   tmp.size) != 0) {
		ret = -EFAULT;
		goto free_newinfo;
	}

	/*
	 * Fix CVE-2016-4997
	 *  -ret = kgr_ip_translate_compat_table(net, tmp.name, tmp.valid_hooks,
	 *  -			     &newinfo, &loc_cpu_entry, tmp.size,
	 *  -			     tmp.num_entries, tmp.hook_entry,
	 *  -			     tmp.underflow);
	 *  +1 line
	 */
	ret = kgr_ip_translate_compat_table(net, &newinfo, &loc_cpu_entry, &tmp);
	if (ret != 0)
		goto free_newinfo;

	duprintf("compat_do_replace: Translated table\n");

	ret = kgr__do_replace(net, tmp.name, tmp.valid_hooks, newinfo,
			   tmp.num_counters, compat_ptr(tmp.counters));
	if (ret)
		goto free_newinfo_untrans;
	return 0;

 free_newinfo_untrans:
	xt_entry_foreach(iter, loc_cpu_entry, newinfo->size)
		kgr_cleanup_entry(iter, net);
 free_newinfo:
	kgr_xt_free_table_info(newinfo);
	return ret;
}

/* patched function */
int kgr_compat_do_ipt_set_ctl(struct sock *sk, int cmd, void __user *user,
		      unsigned int len)
{
	int ret;

	if (!ns_capable(sock_net(sk)->user_ns, CAP_NET_ADMIN))
		return -EPERM;

	switch (cmd) {
	case IPT_SO_SET_REPLACE:
		ret = kgr_compat_do_replace(sock_net(sk), user, len);
		break;

	case IPT_SO_SET_ADD_COUNTERS:
		ret = kgr_do_add_counters(sock_net(sk), user, len, 1);
		break;

	default:
		duprintf("do_ipt_set_ctl:  unknown request %i\n", cmd);
		ret = -EINVAL;
	}

	return ret;
}

static int kgr_patch_bsc986377_kallsyms(void)
{
        unsigned long addr;
        int i;

        for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
                /* mod_find_symname would be nice, but it is not exported */
                addr = kallsyms_lookup_name(kgr_funcs[i].name);
                if (!addr) {
                        pr_err("kgraft-patch: symbol %s not resolved\n",
                                kgr_funcs[i].name);
                        return -ENOENT;
                }

                *(kgr_funcs[i].addr) = (void *)addr;
        }

        return 0;
}

static int kgr_patch_bsc986377_module_notify(struct notifier_block *nb,
        unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc986377_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc986377_module_nb = {
	.notifier_call = kgr_patch_bsc986377_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc986377_ip_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc986377_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc986377_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc986377_ip_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc986377_module_nb);
}
